unit CompForm;

{
  Inno Setup
  Copyright (C) 1998-2002 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  Compiler form

  $Id: CompForm.pas,v 1.26.6.1 2002/02/10 18:35:40 jr Exp $
}

{x$DEFINE STATICCOMPILER}
{ For debugging purposes, remove the 'x' to have it link the compiler code
  into this program and not depend on ISCmplr.dll. }

interface

uses
  WinTypes, WinProcs, SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, Menus, Buttons, ComCtrls,
  SynEdit, SynMemo, SynEditHighlighter, SynHighlighterInno;

const
  CompilerIDEVersion = 'b';

  WM_StartCommandLineCompile = WM_USER + $1000;
  WM_StartCommandLineWizard = WM_USER + $1001;
  WM_ShowStartupForm = WM_USER + $1002;

  MRUListMaxCount = 8;

type
  TCompileForm = class(TForm)
    OpenDialog: TOpenDialog;
    MainMenu1: TMainMenu;
    FMenu: TMenuItem;
    FNew: TMenuItem;
    FOpen: TMenuItem;
    FSave: TMenuItem;
    FSaveAs: TMenuItem;
    N1: TMenuItem;
    FCompile: TMenuItem;
    N2: TMenuItem;
    FExit: TMenuItem;
    EMenu: TMenuItem;
    EUndo: TMenuItem;
    N3: TMenuItem;
    ECut: TMenuItem;
    ECopy: TMenuItem;
    EPaste: TMenuItem;
    EDelete: TMenuItem;
    N4: TMenuItem;
    ESelectAll: TMenuItem;
    VMenu: TMenuItem;
    EFind: TMenuItem;
    EFindNext: TMenuItem;
    EReplace: TMenuItem;
    Help1: TMenuItem;
    HDoc: TMenuItem;
    N6: TMenuItem;
    HAbout: TMenuItem;
    SaveDialog: TSaveDialog;
    FMRUSep: TMenuItem;
    VCompilerStatus: TMenuItem;
    FindDialog: TFindDialog;
    ReplaceDialog: TReplaceDialog;
    StatusPanel: TPanel;
    Status: TListBox;
    SplitPanel: TPanel;
    ToolbarPanel: TPanel;
    NewButton: TSpeedButton;
    Bevel1: TBevel;
    OpenButton: TSpeedButton;
    SaveButton: TSpeedButton;
    CompileButton: TSpeedButton;
    HelpButton: TSpeedButton;
    HWebsite: TMenuItem;
    VToolbar: TMenuItem;
    N7: TMenuItem;
    VOptions: TMenuItem;
    Memo: TSynEdit;
    SynInnoSyn1: TSynInnoSyn;
    PopupMenu1: TPopupMenu;
    PUndo: TMenuItem;
    N8: TMenuItem;
    PCut: TMenuItem;
    PCopy: TMenuItem;
    PPaste: TMenuItem;
    PDelete: TMenuItem;
    PSelectAll: TMenuItem;
    N9: TMenuItem;
    PFind: TMenuItem;
    PFindNext: TMenuItem;
    PReplace: TMenuItem;
    HFaq: TMenuItem;
    Bevel2: TBevel;
    Bevel3: TBevel;
    StatusBar: TStatusBar;
    BodyPanel: TPanel;
    VStatusBar: TMenuItem;
    ERedo: TMenuItem;
    PRedo: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FExitClick(Sender: TObject);
    procedure FOpenClick(Sender: TObject);
    procedure EUndoClick(Sender: TObject);
    procedure EMenuClick(Sender: TObject);
    procedure PopupMenu1Popup(Sender: TObject);
    procedure ECutClick(Sender: TObject);
    procedure ECopyClick(Sender: TObject);
    procedure EPasteClick(Sender: TObject);
    procedure EDeleteClick(Sender: TObject);
    procedure FSaveClick(Sender: TObject);
    procedure ESelectAllClick(Sender: TObject);
    procedure FNewClick(Sender: TObject);
    procedure FNewWizardClick(Sender: TObject);
    procedure FSaveAsClick(Sender: TObject);
    procedure HDocClick(Sender: TObject);
    procedure FCompileClick(Sender: TObject);
    procedure FMenuClick(Sender: TObject);
    procedure FMRUClick(Sender: TObject);
    procedure VCompilerStatusClick(Sender: TObject);
    procedure HAboutClick(Sender: TObject);
    procedure EFindClick(Sender: TObject);
    procedure FindDialogFind(Sender: TObject);
    procedure EReplaceClick(Sender: TObject);
    procedure ReplaceDialogReplace(Sender: TObject);
    procedure EFindNextClick(Sender: TObject);
    procedure SplitPanelMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure VMenuClick(Sender: TObject);
    procedure HWebsiteClick(Sender: TObject);
    procedure VToolbarClick(Sender: TObject);
    procedure VOptionsClick(Sender: TObject);
    procedure HFaqClick(Sender: TObject);
    procedure MemoStatusChange(Sender: TObject;
      Changes: TSynStatusChanges);
    procedure VStatusBarClick(Sender: TObject);
    procedure MemoSpecialLineColors(Sender: TObject; Line: Integer;
      var Special: Boolean; var FG, BG: TColor);
    procedure MemoChange(Sender: TObject);
    procedure MemoDropFiles(Sender: TObject; X, Y: Integer;
      AFiles: TStrings);
    procedure ERedoClick(Sender: TObject);
  private
    { Private declarations }
    Version, Filename: String;
    MRUMenuItems: array[0..MRUListMaxCount-1] of TMenuItem;
    MRUList: TStringList;
    Options: record
      ShowStartupForm: Boolean;
      UseWizard: Boolean;
      TestAfterCompile: Boolean;
      MakeBackups: Boolean;
      FullPathInTitleBar: Boolean;
    end;
    FErrorLine: Integer;
    procedure AddToMRUList (const AFilename: String);
    function AskToSaveModifiedFile: Boolean;
    procedure CompileFile (const AFilename: String;
      const ReadFromFile, PromptToTest: Boolean);
    procedure HideError;
    procedure NewFile;
    procedure NewWizardFile;
    procedure OpenFile (AFilename: String);
    function ProcessPaintMessages: Boolean;
    function SaveFile (const SaveAs: Boolean): Boolean;
    procedure SetErrorLine (ALine: Integer);
    procedure SetStatusPanelVisible (const AVisible: Boolean);
    procedure SetUseSynHigh (const Enable: Boolean);
    procedure StatusMessage (S: String; const AddTime: Boolean);
    procedure UpdateCaption;
    procedure UpdateNewButtons;
    procedure UpdateStatusPanelHeight (H: Integer);
    procedure WMStartCommandLineCompile (var Message: TMessage); message WM_StartCommandLineCompile;
    procedure WMStartCommandLineWizard (var Message: TMessage); message WM_StartCommandLineWizard;
    procedure WMShowStartupForm (var Message: TMessage); message WM_ShowStartupForm;
  public
    { Public declarations }
  end;

var
  CompileForm: TCompileForm;

  CommandLineFilename, CommandLineWizardName: String;
  CommandLineCompile: Boolean;
  CommandLineWizard: Boolean;

procedure InitFormFont (Form: TForm);

implementation

uses
  Clipbrd, ShellApi, IniFiles, Registry, CommDlg,
  MBCSUtil, CmnFunc, CmnFunc2, CompMsgs,
  {$IFDEF STATICCOMPILER} Compile, {$ENDIF}
  CompInt, CompOptions, CompStartup, CompWizard;

{$R *.DFM}

const
  HistoryListSize = 8;

procedure InitFormFont (Form: TForm);
begin
  if DefFontData.Charset = SHIFTJIS_CHARSET then
    Form.Font.Handle := 0  { revert to default Japanese font (requires D3+) }
  else if FontExists('MS Shell Dlg') then
    Form.Font.Name := 'MS Shell Dlg';
end;

function GetDisplayFilename (const Filename: String): String;
var
  Buf: array[0..MAX_PATH-1] of Char;
begin
  if GetFileTitle(PChar(Filename), Buf, SizeOf(Buf)) = 0 then
    Result := Buf
  else
    Result := Filename;
end;


{ TCompileForm }

procedure TCompileForm.FormCreate(Sender: TObject);

  procedure ReadConfig;
  var
    Ini: TRegIniFile;
    I: Integer;
    S: String;
    WindowPlacement: TWindowPlacement;
  begin
    { Don't localize! }
    Ini := TRegIniFile.Create('Software\Jordan Russell\Inno Setup');
    try
      { MRU list }
      for I := 0 to High(MRUMenuItems) do begin
        S := Ini.ReadString('ScriptFileHistoryNew', 'History' + IntToStr(I), '');
        if S <> '' then MRUList.Add (S);
      end;
      { Configuration options }
      ToolbarPanel.Visible := Ini.ReadBool('Options', 'ShowToolbar', True);
      StatusBar.Visible := Ini.ReadBool('Options', 'ShowStatusBar', True);
      Options.ShowStartupForm := Ini.ReadBool('Options', 'ShowStartupForm', True);
      Options.UseWizard := Ini.ReadBool('Options', 'UseWizard', True);
      Options.TestAfterCompile := Ini.ReadBool('Options', 'TestAfterCompile', True);
      Options.MakeBackups := Ini.ReadBool('Options', 'MakeBackups', False);
      Options.FullPathInTitleBar := Ini.ReadBool('Options', 'FullPathInTitleBar', False);
      SetUseSynHigh (Ini.ReadBool('Options', 'UseSynHigh', True));
      if DefFontData.Charset = SHIFTJIS_CHARSET then begin
        { Default to MS Gothic font on a Japanese locale }
        Memo.Font.Name := 'lr SVbN';
        Memo.Font.Size := 9;
        Memo.Font.Charset := SHIFTJIS_CHARSET;
      end;
      Memo.Font.Name := Ini.ReadString('Options', 'EditorFontName', Memo.Font.Name);
      Memo.Font.Size := Ini.ReadInteger('Options', 'EditorFontSize', Memo.Font.Size);
      Memo.Font.Charset := Ini.ReadInteger('Options', 'EditorFontCharset', Memo.Font.Charset);
      UpdateNewButtons;
      { Window state }
      if Ini.ReadBool('State', 'WindowMaximized', False) then
        WindowState := wsMaximized;
      WindowPlacement.length := SizeOf(WindowPlacement);
      GetWindowPlacement (Handle, @WindowPlacement);
      WindowPlacement.rcNormalPosition.Left := Ini.ReadInteger('State',
        'WindowLeft', WindowPlacement.rcNormalPosition.Left);
      WindowPlacement.rcNormalPosition.Top := Ini.ReadInteger('State',
        'WindowTop', WindowPlacement.rcNormalPosition.Top);
      WindowPlacement.rcNormalPosition.Right := Ini.ReadInteger('State',
        'WindowRight', WindowPlacement.rcNormalPosition.Right);
      WindowPlacement.rcNormalPosition.Bottom := Ini.ReadInteger('State',
        'WindowBottom', WindowPlacement.rcNormalPosition.Bottom);
      SetWindowPlacement (Handle, @WindowPlacement);
    finally
      Ini.Free;
    end;
  end;

var
  I: Integer;
  NewItem: TMenuItem;
begin
  {$IFNDEF STATICCOMPILER}
  Version := ISDllGetVersion.Version;
  {$ELSE}
  Version := ISGetVersion.Version;
  {$ENDIF}

  InitFormFont (Self);

  { For some reason, if AutoScroll=False is set on the form Delphi ignores the
    'poDefault' Position setting }
  AutoScroll := False;

  { Append 'Del' to the end of the Delete item. Don't actually use Del as
    the shortcut key so that the Del key still works when the menu item is
    disabled because there is no selection. }
  EDelete.Caption := EDelete.Caption + #9 + ShortCutToText(VK_DELETE);

  MRUList := TStringList.Create;
  for I := 0 to High(MRUMenuItems) do begin
    NewItem := TMenuItem.Create(Self);
    NewItem.OnClick := FMRUClick;
    FMenu.Insert (FMenu.IndexOf(FMRUSep), NewItem);
    MRUMenuItems[I] := NewItem;
  end;

  UpdateCaption;
  OpenDialog.Filter := SCompilerOpenFilter;
  SaveDialog.Filter := SCompilerOpenFilter;

  if CommandLineCompile then
    PostMessage (Handle, WM_StartCommandLineCompile, 0, 0)
  else if CommandLineWizard then begin
    { Stop Delphi from showing the compiler form }
    Application.ShowMainForm := False;
    { Show wizard form later }
    PostMessage (Handle, WM_StartCommandLineWizard, 0, 0);
  end else begin
    ReadConfig;
    if CommandLineFilename = '' then begin
      if Options.ShowStartupForm then
        PostMessage (Handle, WM_ShowStartupForm, 0, 0);
    end else
      OpenFile (CommandLineFilename)
  end;
end;

procedure TCompileForm.FormDestroy(Sender: TObject);

  procedure SaveConfig;
  var
    Ini: TRegIniFile;
    I: Integer;
    S: String;
    WindowPlacement: TWindowPlacement;
  begin
    { Don't localize! }
    Ini := TRegIniFile.Create('Software\Jordan Russell\Inno Setup');
    try
      { MRU list }
      for I := 0 to High(MRUMenuItems) do begin
        if I < MRUList.Count then
          S := MRUList[I]
        else
          S := '';
        Ini.WriteString ('ScriptFileHistoryNew', 'History' + IntToStr(I),
          S {work around Delphi 2 bug:} + #0);
      end;
      { Configuration options }
      Ini.WriteBool ('Options', 'ShowToolbar', ToolbarPanel.Visible);
      Ini.WriteBool ('Options', 'ShowStatusBar', StatusBar.Visible);
      Ini.WriteBool ('Options', 'ShowStartupForm', Options.ShowStartupForm);
      Ini.WriteBool ('Options', 'UseWizard', Options.UseWizard);
      Ini.WriteBool ('Options', 'TestAfterCompile', Options.TestAfterCompile);
      Ini.WriteBool ('Options', 'MakeBackups', Options.MakeBackups);
      Ini.WriteBool ('Options', 'FullPathInTitleBar', Options.FullPathInTitleBar);
      Ini.WriteBool ('Options', 'UseSynHigh', Assigned(Memo.Highlighter));
      Ini.WriteString ('Options', 'EditorFontName', Memo.Font.Name);
      Ini.WriteInteger ('Options', 'EditorFontSize', Memo.Font.Size);
      Ini.WriteInteger ('Options', 'EditorFontCharset', Memo.Font.Charset);
      { Window state }
      WindowPlacement.length := SizeOf(WindowPlacement);
      GetWindowPlacement (Handle, @WindowPlacement);
      Ini.WriteInteger ('State', 'WindowLeft', WindowPlacement.rcNormalPosition.Left);
      Ini.WriteInteger ('State', 'WindowTop', WindowPlacement.rcNormalPosition.Top);
      Ini.WriteInteger ('State', 'WindowRight', WindowPlacement.rcNormalPosition.Right);
      Ini.WriteInteger ('State', 'WindowBottom', WindowPlacement.rcNormalPosition.Bottom);
      Ini.WriteBool ('State', 'WindowMaximized', WindowState = wsMaximized);
    finally
      Ini.Free;
    end;
  end;

begin
  if not (CommandLineCompile or CommandLineWizard) then
    SaveConfig;

  MRUList.Free;
end;

procedure TCompileForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  CanClose := AskToSaveModifiedFile;
end;

procedure TCompileForm.UpdateCaption;
var
  NewCaption: String;
begin
  if Filename = '' then
    NewCaption := 'Untitled'
  else begin
    if Options.FullPathInTitleBar then
      NewCaption := Filename
    else
      NewCaption := GetDisplayFilename(Filename);
  end;
  NewCaption := NewCaption + ' - ' + SCompilerFormCaption + ' ' + Version;
  Caption := NewCaption;
  if not CommandLineWizard then
    Application.Title := NewCaption;
end;

procedure TCompileForm.UpdateNewButtons;
begin
  if Options.UseWizard then begin
    FNew.OnClick := FNewWizardClick;
    NewButton.OnClick := FNewWizardClick;
  end else begin
    FNew.OnClick := FNewClick;
    NewButton.OnClick := FNewClick;
  end;
end;

procedure TCompileForm.NewFile;
begin
  Memo.Text := '';
  Memo.Modified := False;
  Filename := '';
  UpdateCaption;
  HideError;
end;

procedure TCompileForm.NewWizardFile;
var
  WizardForm: TWizardForm;
  F: TextFile;
begin
  WizardForm := TWizardForm.Create(Application);
  try
    if CommandLineWizard then
      WizardForm.WizardName := CommandLineWizardName;

    if WizardForm.ShowModal <> mrOk then
      Exit;

    if CommandLineWizard then begin
      AssignFile(F, CommandLineFileName);
      FileMode := fmOpenWrite or fmShareExclusive;
      Rewrite(F);
      try
        Write(F, WizardForm.ResultScript);
      finally
        CloseFile(F);
      end;
    end else begin
      Memo.Text := WizardForm.ResultScript;
      Memo.Modified := (WizardForm.Result = wrComplete);
      Filename := '';
      UpdateCaption;
      HideError;

      if WizardForm.Result = wrComplete then
        if MsgBox('Would you like to compile the new script now?', SCompilerFormCaption, mbConfirmation, MB_YESNO) = ID_YES then
          FCompileClick(Self);
    end;
  finally
    WizardForm.Free;
  end;
end;

procedure TCompileForm.OpenFile (AFilename: String);
begin
  AFilename := ExpandFileName(AFilename);
  AddToMRUList (AFilename);
  Memo.Lines.LoadFromFile (AFilename);
  Memo.Modified := False;
  Filename := AFilename;
  UpdateCaption;
  HideError;
end;

function TCompileForm.SaveFile (const SaveAs: Boolean): Boolean;

  procedure SaveTo (const FN: String);
  var
    BackupFN: String;
  begin
    if Options.MakeBackups and NewFileExists(FN) then begin
      BackupFN := ChangeFileExt(FN, '.~is');
      DeleteFile (BackupFN);
      if not RenameFile(FN, BackupFN) then
        raise Exception.Create('Error creating backup file. Could not save file');
    end;
    Memo.Lines.SaveToFile (FN);
  end;

begin
  Result := False;
  if SaveAs or (Filename = '') then begin
    SaveDialog.Filename := Filename;
    if not SaveDialog.Execute then Exit;
    SaveTo (SaveDialog.Filename);
    Filename := SaveDialog.Filename;
    UpdateCaption;
  end
  else
    SaveTo (Filename);
  Memo.Modified := False;
  Memo.ClearUndo;
  Result := True;
  AddToMRUList (Filename);
end;

function TCompileForm.AskToSaveModifiedFile: Boolean;
var
  FileTitle: String;
begin
  Result := True;
  if Memo.Modified then begin
    FileTitle := Filename;
    if FileTitle = '' then FileTitle := 'Untitled';
    case MsgBox('The text in the ' + FileTitle + ' file has changed.'#13#10#13#10 +
       'Do you want to save the changes?', SCompilerFormCaption, mbError,
       MB_YESNOCANCEL) of
      ID_YES: Result := SaveFile(False);
      ID_NO: ;
    else
      Result := False;
    end;
  end;
end;

procedure TCompileForm.AddToMRUList (const AFilename: String);
var
  I: Integer;
begin
  I := 0;
  while I < MRUList.Count do begin
    if AnsiCompareFileName(MRUList[I], AFilename) = 0 then
      MRUList.Delete (I)
    else
      Inc (I);
  end;
  MRUList.Insert (0, AFilename);
  while MRUList.Count > High(MRUMenuItems)+1 do
    MRUList.Delete (MRUList.Count-1);
end;

procedure TCompileForm.StatusMessage (S: String; const AddTime: Boolean);
var
  DC: HDC;
  Size: TSize;
begin
  if AddTime then
    S := S + '  [' + TimeToStr(Time) + ']';
  with Status do begin
    try
      TopIndex := Items.Add(S);
    except
      on EOutOfResources do begin
        Clear;
        SendMessage (Handle, LB_SETHORIZONTALEXTENT, 0, 0);
        Items.Add (SCompilerStatusReset);
        TopIndex := Items.Add(S);
      end;
    end;
    DC := GetDC(0);
    try
      SelectObject (DC, Font.Handle);
      GetTextExtentPoint (DC, PChar(S), Length(S), Size);
    finally
      ReleaseDC (0, DC);
    end;
    Inc (Size.cx, 5);
    if Size.cx > SendMessage(Handle, LB_GETHORIZONTALEXTENT, 0, 0) then
      SendMessage (Handle, LB_SETHORIZONTALEXTENT, Size.cx, 0);
  end;
end;

type
  PAppData = ^TAppData;
  TAppData = record
    Form: TCompileForm;
    ScriptFile: ^TextFile;
    CurLineNumber: Integer;
    CurLine: String;
    OutputExe: String;
    ErrorMsg: String;
    ErrorFilename: String;
    ErrorLine: Integer;
    Aborted: Boolean;
  end;

function CompilerCallbackProc (Code: Integer; var Data: TCompilerCallbackData;
  AppData: Longint): Integer; stdcall;
begin
  Result := iscrSuccess;
  with PAppData(AppData)^ do
    case Code of
      iscbReadScript:
        if ScriptFile = nil then begin
          if Data.Reset then
            CurLineNumber := 0;
          if CurLineNumber < Form.Memo.Lines.Count then begin
            CurLine := Form.Memo.Lines[CurLineNumber];
            Data.LineRead := PChar(CurLine);
            Inc (CurLineNumber);
          end;
        end
        else begin
          if Data.Reset then
            Reset (ScriptFile^);
          if not Eof(ScriptFile^) then begin
            Readln (ScriptFile^, CurLine);
            Data.LineRead := PChar(CurLine);
          end;
        end;
      iscbNotifyStatus:
        Form.StatusMessage (Data.StatusMsg, False);
      iscbNotifyIdle:
        if Form.ProcessPaintMessages then
          Result := iscrRequestAbort;
      iscbNotifySuccess:
        OutputExe := Data.OutputExeFilename;
      iscbNotifyError:
        begin
          if Assigned(Data.ErrorMsg) then
            ErrorMsg := Data.ErrorMsg
          else
            Aborted := True;
          ErrorFilename := Data.ErrorFilename;
          ErrorLine := Data.ErrorLine;
        end;
    end;
end;

function TCompileForm.ProcessPaintMessages: Boolean;
{ Dispatches all pending WM_PAINT messages. In effect, this is like an
  'UpdateWindow' on all visible windows }
var
  Msg: TMsg;
begin
  Result := False;
  while PeekMessage(Msg, 0, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE) do begin
    if Msg.message = WM_QUIT then begin
      { Repost WM_QUIT messages }
      PostQuitMessage (Msg.WParam);
      Exit;
    end;
    if (Msg.message = WM_KEYDOWN) and (Msg.wParam = VK_ESCAPE) then begin
      Result := True;
      Break;
    end;
  end;
  while PeekMessage(Msg, 0, WM_PAINT, WM_PAINT, PM_NOREMOVE) do begin
    case Integer(GetMessage(Msg, 0, WM_PAINT, WM_PAINT)) of
      -1: Break; { if GetMessage failed }
      0: begin
           { Repost WM_QUIT messages }
           PostQuitMessage (Msg.WParam);
           Break;
         end;
    end;
    DispatchMessage (Msg);
  end;
end;

procedure TCompileForm.CompileFile (const AFilename: String;
  const ReadFromFile, PromptToTest: Boolean);
var
  F: TextFile;
  SourcePath, S: String;
  Params: TCompileScriptParams;
  AppData: TAppData;
begin
  if ReadFromFile then begin
    AssignFile (F, AFilename);
    FileMode := fmOpenRead or fmShareDenyWrite;  Reset (F);
  end;
  try
    Screen.Cursor := crHourglass;
    Enabled := False;
    HideError;
    Status.Clear;
    SendMessage (Status.Handle, LB_SETHORIZONTALEXTENT, 0, 0);
    SetStatusPanelVisible (True);

    SourcePath := ExtractFilePath(AFilename);
    FillChar (Params, SizeOf(Params), 0);
    Params.Size := SizeOf(Params);
    Params.CompilerPath := nil;
    Params.SourcePath := PChar(SourcePath);
    Params.CallbackProc := CompilerCallbackProc;
    Pointer(Params.AppData) := @AppData;

    AppData.Form := Self;
    if ReadFromFile then
      AppData.ScriptFile := @F
    else
      AppData.ScriptFile := nil;
    AppData.CurLineNumber := 0;
    AppData.Aborted := False;

    StatusMessage (SCompilerStatusStarting, True);
    StatusMessage ('', False);
    {$IFNDEF STATICCOMPILER}
    if ISDllCompileScript(Params) <> isceNoError then begin
    {$ELSE}
    if ISCompileScript(Params, False) <> isceNoError then begin
    {$ENDIF}
      StatusMessage (SCompilerStatusErrorAborted, False);
      Enabled := True;
      if (AppData.ScriptFile = nil) and (AppData.ErrorLine > 0) then begin
        { Move the caret to the line number the error occured on }
        if Memo.CaretY <> AppData.ErrorLine then
          Memo.CaretXY := Point(1, AppData.ErrorLine);
        Memo.SetFocus;
        SetErrorLine (AppData.ErrorLine);
      end;
      if not AppData.Aborted then begin
        S := '';
        if AppData.ErrorFilename <> '' then
          S := 'File: ' + AppData.ErrorFilename + SNewLine2;
        if AppData.ErrorLine > 0 then
          S := S + Format('Line %d:' + SNewLine, [AppData.ErrorLine]);
        S := S + AppData.ErrorMsg;
        MsgBox (S, 'Compiler Error', mbCriticalError, MB_OK)
      end
      else
        MsgBox ('Compile aborted.', 'Compiler', mbError, MB_OK);
      Exit;
    end;
    StatusMessage (SCompilerStatusFinished, True);

    Enabled := True;
    if PromptToTest and Options.TestAfterCompile then begin
      if MsgBoxFmt(SCompilerSuccessfulMessage2,
         [RemoveBackslashUnlessRoot(ExtractFilePath(AppData.OutputExe))],
         SCompilerSuccessfulTitle, mbInformation,
         MB_YESNO or MB_DEFBUTTON2) = ID_YES then
        if ShellExecute(Handle, nil, StringAsPChar(AppData.OutputExe), '',
           StringAsPChar(AppData.OutputExe), SW_SHOW) <= 32 then
          raise Exception.Create(SCompilerExecuteSetupError);
    end;
  finally
    Enabled := True;
    Screen.Cursor := crDefault;
    if ReadFromFile then
      CloseFile (F);
  end;
end;

procedure TCompileForm.SetUseSynHigh (const Enable: Boolean);
begin
  if Enable then
    Memo.Highlighter := SynInnoSyn1
  else
    Memo.Highlighter := nil;
end;

procedure TCompileForm.FMenuClick(Sender: TObject);
var
  I: Integer;
begin
  FMRUSep.Visible := MRUList.Count <> 0;
  for I := 0 to High(MRUMenuItems) do
    with MRUMenuItems[I] do begin
      if I < MRUList.Count then begin
        Visible := True;
        Caption := '&' + IntToStr(I+1) + ' ' + MRUList[I];
      end
      else
        Visible := False;
    end;
end;

procedure TCompileForm.FNewClick(Sender: TObject);
begin
  if not AskToSaveModifiedFile then Exit;
  NewFile;
end;

procedure TCompileForm.FNewWizardClick(Sender: TObject);
begin
  if not AskToSaveModifiedFile then Exit;
  NewWizardFile;
end;

procedure TCompileForm.FOpenClick(Sender: TObject);
begin
  OpenDialog.Filename := '';
  if not AskToSaveModifiedFile or not OpenDialog.Execute then
    Exit;
  OpenFile (OpenDialog.Filename);
end;

procedure TCompileForm.FSaveClick(Sender: TObject);
begin
  SaveFile (False);
end;

procedure TCompileForm.FSaveAsClick(Sender: TObject);
begin
  SaveFile (True);
end;

procedure TCompileForm.FCompileClick(Sender: TObject);
begin
  if Filename = '' then begin
    if MsgBox('The file must be saved before compiling.'#13#10#13#10 +
       'Do you want to save it now?', SCompilerFormCaption, mbError,
       MB_YESNO) <> ID_YES then
      Exit;
    if not SaveFile(False) then
      Exit;
  end;
  CompileFile (Filename, False, True);
end;

procedure TCompileForm.FMRUClick(Sender: TObject);
var
  I: Integer;
begin
  if not AskToSaveModifiedFile then Exit;
  for I := 0 to High(MRUMenuItems) do
    if MRUMenuItems[I] = Sender then begin
      OpenFile (MRUList[I]);
      Break;
    end;
end;

procedure TCompileForm.FExitClick(Sender: TObject);
begin
  Close;
end;

procedure TCompileForm.EMenuClick(Sender: TObject);
begin
  EUndo.Enabled := Memo.CanUndo;
  ERedo.Enabled := Memo.CanRedo;
  ECut.Enabled := Memo.SelAvail;
  ECopy.Enabled := Memo.SelAvail;
  EPaste.Enabled := Clipboard.HasFormat(CF_TEXT);
  EDelete.Enabled := Memo.SelAvail;
end;

procedure TCompileForm.PopupMenu1Popup(Sender: TObject);
begin
  PUndo.Enabled := Memo.CanUndo;
  PRedo.Enabled := Memo.CanRedo;
  PCut.Enabled := Memo.SelAvail;
  PCopy.Enabled := Memo.SelAvail;
  PPaste.Enabled := Clipboard.HasFormat(CF_TEXT);
  PDelete.Enabled := Memo.SelAvail;
end;

procedure TCompileForm.EUndoClick(Sender: TObject);
begin
  Memo.Undo;
end;

procedure TCompileForm.ERedoClick(Sender: TObject);
begin
  Memo.Redo;
end;

procedure TCompileForm.ECutClick(Sender: TObject);
begin
  Memo.CutToClipboard;
end;

procedure TCompileForm.ECopyClick(Sender: TObject);
begin
  Memo.CopyToClipboard;
end;

procedure TCompileForm.EPasteClick(Sender: TObject);
begin
  Memo.PasteFromClipboard;
end;

procedure TCompileForm.EDeleteClick(Sender: TObject);
begin
  Memo.ClearSelection;
end;

procedure TCompileForm.ESelectAllClick(Sender: TObject);
begin
  Memo.SelectAll;
end;

procedure TCompileForm.VMenuClick(Sender: TObject);
begin
  VToolbar.Checked := ToolbarPanel.Visible;
  VStatusBar.Checked := StatusBar.Visible;
  VCompilerStatus.Checked := StatusPanel.Visible;
end;

procedure TCompileForm.VToolbarClick(Sender: TObject);
begin
  ToolbarPanel.Visible := not ToolbarPanel.Visible;
end;

procedure TCompileForm.VStatusBarClick(Sender: TObject);
begin
  StatusBar.Visible := not StatusBar.Visible;
end;

procedure TCompileForm.SetStatusPanelVisible (const AVisible: Boolean);
begin
  if StatusPanel.Visible <> AVisible then begin
    SplitPanel.Visible := AVisible;
    StatusPanel.Visible := AVisible;
    if AVisible then
      StatusPanel.Top := ClientHeight;
  end;
end;

procedure TCompileForm.VCompilerStatusClick(Sender: TObject);
begin
  SetStatusPanelVisible (not StatusPanel.Visible);
end;

procedure TCompileForm.HDocClick(Sender: TObject);
begin
  Application.HelpCommand (HELP_FINDER, 0);
end;

procedure TCompileForm.HFaqClick(Sender: TObject);
begin
  ShellExecute (Application.Handle, 'open',
    PChar(ExtractFilePath(ParamStr(0)) + 'isfaq.htm'), nil, nil, SW_SHOW);
end;

procedure TCompileForm.HWebsiteClick(Sender: TObject);
begin
  ShellExecute (Application.Handle, 'open', 'http://www.innosetup.com/', nil,
    nil, SW_SHOW);
end;

procedure TCompileForm.HAboutClick(Sender: TObject);
begin
  { Removing the About box or modifying the text inside it is a violation of the
    Inno Setup license agreement; see LICENSE.TXT. However, adding additional
    lines to the About box is permitted. }
  MsgBox ('Inno Setup Compiler version ' +
    Version + ' (' + CompilerIDEVersion + ')' + SNewLine +
    'Copyright (C) 1998-2002 Jordan Russell. All rights reserved.' + SNewLine +
    'Portions by Martijn Laan.' + SNewLine2 +
    'Inno Setup home page:' + SNewLine +
    'http://www.innosetup.com/' + SNewLine2 +
    'Refer to LICENSE.TXT for conditions of distribution and use.',
    'About Inno Setup', mbInformation, MB_OK);
end;

procedure TCompileForm.WMStartCommandLineCompile (var Message: TMessage);
var
  Code: Integer;
begin
  UpdateStatusPanelHeight (ClientHeight);
  Code := 0;
  try
    try
      CompileFile (CommandLineFilename, True, False);
    except
      on E: Exception do begin
        Code := 2;
        Application.HandleException (E);
      end;
    end;
  finally
    Halt (Code);
  end;
end;

procedure TCompileForm.WMStartCommandLineWizard (var Message: TMessage);
var
  Code: Integer;
begin
  Code := 0;
  try
    try
      NewWizardFile;
    except
      on E: Exception do begin
        Code := 2;
        Application.HandleException (E);
      end;
    end;
  finally
    Halt(Code);
  end;
end;

procedure TCompileForm.WMShowStartupForm (var Message: TMessage);
var
  StartupForm: TStartupForm;
begin
  StartupForm := TStartupForm.Create(Application);
  try
    StartupForm.MRUList := MRUList;
    StartupForm.StartupCheck.Checked := not Options.ShowStartupForm;
    if StartupForm.ShowModal = mrOK then begin
      Options.ShowStartupForm := not StartupForm.StartupCheck.Checked;
      case StartupForm.Result of
        srEmpty:
          FNewClick(Self);
        srWizard:
          FNewWizardClick(Self);
        srOpenFile:
          if AskToSaveModifiedFile then
            OpenFile(StartupForm.ResultFileName);
        srOpenDialog:
          FOpenClick(Self);
      end;
    end;
  finally
    StartupForm.Free;
  end;
end;

procedure TCompileForm.EFindClick(Sender: TObject);
begin
  ReplaceDialog.CloseDialog;
  FindDialog.Execute;
end;

procedure TCompileForm.EFindNextClick(Sender: TObject);
begin
  if FindDialog.FindText = '' then
    EFindClick (Sender)
  else
    FindDialogFind (FindDialog);
end;

function FindOptionsToSearchOptions (const FindOptions: TFindOptions): TSynSearchOptions;
begin
  Result := [];
  if frMatchCase in FindOptions then
    Include (Result, ssoMatchCase);
  if frWholeWord in FindOptions then
    Include (Result, ssoWholeWord);
  if not(frDown in FindOptions) then
    Include (Result, ssoBackwards);
end;

procedure TCompileForm.FindDialogFind(Sender: TObject);
begin
  { this event handler is shared between FindDialog & ReplaceDialog }
  with Sender as TFindDialog do
    if Memo.SearchReplace(FindText, '', FindOptionsToSearchOptions(Options)) = 0 then
      MsgBoxFmt ('Cannot find "%s"', [FindText], '', mbError, MB_OK);
end;

procedure TCompileForm.EReplaceClick(Sender: TObject);
begin
  FindDialog.CloseDialog;
  ReplaceDialog.Execute;
end;

procedure TCompileForm.ReplaceDialogReplace(Sender: TObject);
var
  Same: Boolean;
begin
  with ReplaceDialog do begin
    if not(frMatchCase in Options) then
      Same := AnsiCompareText(Memo.SelText, FindText) = 0
    else
      Same := Memo.SelText = FindText;
    if Same then
      Memo.SelText := ReplaceText;
    if Memo.SearchReplace(FindText, '', FindOptionsToSearchOptions(Options)) = 0 then
      MsgBoxFmt ('Cannot find "%s"', [FindText], '', mbError, MB_OK)
    else
      if frReplaceAll in Options then begin
        repeat
          Memo.SelText := ReplaceText;
        until Memo.SearchReplace(FindText, '', FindOptionsToSearchOptions(Options)) = 0;
      end;
  end;
end;

procedure TCompileForm.UpdateStatusPanelHeight (H: Integer);
var
  H2: Integer;
begin
  H2 := BodyPanel.ClientHeight - 48 - SplitPanel.Height;
  if H > H2 then H := H2;
  if H < 48 then H := 48;
  StatusPanel.Height := H;
end;

procedure TCompileForm.SplitPanelMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  P: TPoint;
begin
  if (ssLeft in Shift) and StatusPanel.Visible then begin
    GetCursorPos (P);
    UpdateStatusPanelHeight (BodyPanel.ClientToScreen(Point(0, 0)).Y - P.Y +
      BodyPanel.ClientHeight - (SplitPanel.Height div 2));
  end;
end;

procedure TCompileForm.VOptionsClick(Sender: TObject);
var
  OptionsForm: TOptionsForm;
begin
  OptionsForm := TOptionsForm.Create(Application);
  try
    OptionsForm.StartupCheck.Checked := Options.ShowStartupForm;
    OptionsForm.WizardCheck.Checked := Options.UseWizard;
    OptionsForm.TestCheck.Checked := Options.TestAfterCompile;
    OptionsForm.BackupCheck.Checked := Options.MakeBackups;
    OptionsForm.FullPathCheck.Checked := Options.FullPathInTitleBar;
    OptionsForm.UseSynHighCheck.Checked := Assigned(Memo.Highlighter);
    OptionsForm.FontPanel.Font.Assign (Memo.Font);

    if OptionsForm.ShowModal <> mrOK then
      Exit;

    Options.ShowStartupForm := OptionsForm.StartupCheck.Checked;
    Options.UseWizard := OptionsForm.WizardCheck.Checked;
    Options.TestAfterCompile := OptionsForm.TestCheck.Checked;
    Options.MakeBackups := OptionsForm.BackupCheck.Checked;
    Options.FullPathInTitleBar := OptionsForm.FullPathCheck.Checked;
    UpdateCaption;
    SetUseSynHigh (OptionsForm.UseSynHighCheck.Checked);
    Memo.Font.Assign (OptionsForm.FontPanel.Font);
    UpdateNewButtons;
  finally
    OptionsForm.Free;
  end;
end;

procedure TCompileForm.SetErrorLine (ALine: Integer);
begin
  if FErrorLine <> ALine then begin
    if FErrorLine > 0 then
      Memo.InvalidateLine (FErrorLine);
    FErrorLine := ALine;
    if FErrorLine > 0 then
      Memo.InvalidateLine (FErrorLine);
  end;
end;

procedure TCompileForm.HideError;
begin
  SetErrorLine (0);
end;

procedure TCompileForm.MemoStatusChange(Sender: TObject;
  Changes: TSynStatusChanges);
const
  InsertText: array[Boolean] of String = ('Overwrite', 'Insert');
begin
  if (scCaretX in Changes) or (scCaretY in Changes) then begin
    HideError;
    StatusBar.Panels[0].Text := Format('%4d:%4d', [Memo.CaretY, Memo.CaretX]);
  end;
  if scModified in Changes then begin
    if Memo.Modified then
      StatusBar.Panels[1].Text := 'Modified'
    else
      StatusBar.Panels[1].Text := '';
  end;
  if scInsertMode in Changes then
    StatusBar.Panels[2].Text := InsertText[Memo.InsertMode];
end;

procedure TCompileForm.MemoChange(Sender: TObject);
begin
  { Need HideError here because we don't get an OnStatusChange event when the
    Delete key is pressed }
  HideError;
end;

procedure TCompileForm.MemoSpecialLineColors(Sender: TObject;
  Line: Integer; var Special: Boolean; var FG, BG: TColor);
begin
  if FErrorLine = Line then begin
    Special := True;
    FG := clWhite;
    BG := clMaroon;
  end;
end;

procedure TCompileForm.MemoDropFiles(Sender: TObject; X, Y: Integer;
  AFiles: TStrings);
begin
  if (AFiles.Count > 0) and AskToSaveModifiedFile then
    OpenFile (AFiles[0]);
end;

end.
