unit WizardClassic;

{
  Inno Setup
  Copyright (C) 1998-2000 Jordan Russell
  Portions by Martijn Laan
  For conditions of distribution and use, see LICENSE.TXT.

  Wizard form

  $Id: WizardClassic.pas,v 1.8 2001/04/19 16:49:40 jr Exp $
}

interface

{$IFNDEF WIZARDCLASSIC}
  Please define WIZARDCLASSIC
{$ENDIF}

uses
  WinTypes, WinProcs, SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, NewBevel, NewFCtrl, ExtCtrls,
  ChildFrm, Struct, NewCheckListBox, RichEditViewer, BeveledLabel, NewGauge;

type
  TWizardPage = (wpWelcome, wpLicense, wpPassword, wpInfoBefore, wpSelectDir,
    wpSelectComponents, wpSelectProgramGroup, wpSelectTasks, wpReady,
    wpInstalling, wpInfoAfter, wpFinished);

  TWizardForm = class(TSetupChildForm)
    CancelButton: TButton;
    NextButton: TButton;
    BackButton: TButton;
    Notebook1: TNotebook;
    PicturePaintBox: TPaintBox;
    Bevel: TBevel;
    Notebook2: TNotebook;
    WelcomeLabel: TLabel;
    SelectDirLabel: TLabel;
    DiskSpaceLabel: TLabel;
    DirEdit: TEdit;
    DirList: TNewDirectoryListBox;
    DriveList: TNewDriveComboBox;
    SelectStartMenuFolderLabel: TLabel;
    GroupEdit: TEdit;
    NoIconsCheck: TCheckBox;
    GroupList: TListBox;
    LicenseLabel1: TLabel;
    LicenseMemo: TRichEditViewer;
    LicenseLabel2: TLabel;
    LicenseIcon: TPaintBox;
    InfoBeforeIcon: TPaintBox;
    InfoBeforeLabel: TLabel;
    InfoBeforeMemo: TRichEditViewer;
    InfoAfterIcon: TPaintBox;
    InfoAfterLabel: TLabel;
    InfoAfterMemo: TRichEditViewer;
    FinishedLabel: TLabel;
    InfoBeforeClickLabel: TLabel;
    InfoAfterClickLabel: TLabel;
    YesRadio: TRadioButton;
    NoRadio: TRadioButton;
    ReadyLabel: TLabel;
    PasswordLabel: TLabel;
    PasswordEdit: TEdit;
    PasswordEditLabel: TLabel;
    SelectComponentsLabel: TLabel;
    ReadyMemo: TMemo;
    TypesCombo: TComboBox;
    ComponentsList: TNewCheckListBox;
    SelectTasksLabel: TLabel;
    BaseTaskLabel: TLabel;
    BaseTaskCheckBox: TCheckBox;
    ComponentsDiskSpaceLabel: TLabel;
    BeveledLabel: TBeveledLabel;
    StatusLabel: TLabel;
    FilenameLabel: TNewPathLabel;
    ProgressGauge: TNewGauge;
    InstallingLabel: TLabel;
    procedure NextButtonClick(Sender: TObject);
    procedure BackButtonClick(Sender: TObject);
    procedure CancelButtonClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure DirListChange(Sender: TObject);
    procedure GroupListClick(Sender: TObject);
    procedure NoIconsCheckClick(Sender: TObject);
    procedure PicturePaintBoxPaint(Sender: TObject);
    procedure GroupEditChange(Sender: TObject);
    procedure SetupIconPaint(Sender: TObject);
    procedure TypesComboChange(Sender: TObject);
    procedure ComponentsListClickCheck(Sender: TObject);
    procedure ReadyMemoKeyPress(Sender: TObject; var Key: Char);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    OldTextHeight, NewTextHeight: Integer;
    NeedSelectTasksPageUpdate: Boolean;
    FirstRunButton: TButtonControl;
    TaskLabels, TaskButtons: TList;
    PreviousAppDir: String;
    PrevSelectedComponents, PrevDeselectedComponents: TStringList;
    PrevSelectedTasks, PrevDeselectedTasks: TStringList;
    InitialSelectedComponents: TStringList;
    InitialSetupTypeIndex: Integer;
    procedure CalcComponentsSpace;
    function CreateCheckBox(const BaseControl: TControl; const ACaption: String;
      const AChecked: Boolean; const ATag: Integer;
      var Y: Integer): TButtonControl;
    function CreateLabel(const BaseControl: TControl; const ACaption: String;
      var Y: Integer): TLabel;
    function CreateRadioButton(const BaseControl: TControl;
      const ACaption, AGroup: String; const AChecked: Boolean;
      const ATag: Integer; var Y: Integer): TButtonControl;
    procedure CreateTaskButtons(const SelectedComponents: TStringList);
    function GetFirstTaskButton: TButtonControl;
    procedure ReadGroupList;
    procedure WMSysCommand (var Message: TWMSysCommand); message WM_SYSCOMMAND;
  public
    { Public declarations }
    CurPage: TWizardPage;
    OrigFinishedLabelWidth: Integer;
    constructor Create (AOwner: TComponent); override;
    procedure ChangeFinishedLabel (const S: String);
    procedure ChangeReadyLabel (const S: String);
    procedure CreateRunButtons(const SelectedComponents, SelectedTasks: TStringList; RunButtons: TList);
    function GetSetupType: PSetupTypeEntry;
    procedure SetSelectedComponents(const SelectedComponents, DeselectedComponents: TStringList);
    procedure SetSelectedComponentsFromType(const TypeName: String; OnlySelectFixedComponents: Boolean);
    procedure GetSelectedComponents(Components: TStringList; const Descriptions: Boolean);
    procedure UpdateSelectTasksPage;
    procedure GetSelectedTasks(Tasks: TStringList; const Descriptions: Boolean);
    procedure CurPageChanged;
    function SkipCurPage: Boolean;
  end;

var
  WizardForm: TWizardForm;

function FixLabel (const S: String; const Space: LongInt): String;

implementation

uses
  ShlObj, GroupRadioButton,
  Msgs, MsgIDs, Main, Install, MBCSUtil, CmnFunc, CmnFunc2,
  DDEInt, zlib;

{$R *.DFM}

{$I Wizard.inc}

end.
